/*
Code to demonstrate a serious data corruption bug in PVCam.

The bug is that one can reliably get corrupted images (a band of zeros of varying width along the top) as follows:
- reset the camera:
   - unload device driver
   - power off camera
   - delete device
   - power on camera
   - open the camera
- take 3 full-frame unbinned 0-second exposures using a circular buffer and internal timing
- take 3 binned 2x3 1-second exposures using a circular buffer and external sync pulse for timing
- the 2nd set of exposures are all corrupted
The only workaround I have found is to close and reopen the camera between every sequence of exposures.

We see this bug with the following setup:
- Princeton Instruments VersArray 1024 frame-transfer camera with a USB interface
- Princeton Instruments PVCam library (we have version 2.7.1.7)
- Linux (we see this on Fedora Core 4)

Requirements:
- Run a square wave of some suitable frequency (slower than the camera's readout time!)
  into the sync pulse input of the camera controller.
- cfitsio (a free package), or if you prefer you can comment out the fits writing portion of the code
  and simply test that the upper left pixel is 0 or write the data out in some other format.
*/
// uncomment the following line if you are driving the sync pulse with PCI-SYNCCLOCK card and our software to control it
//#define HAVESYNCCLOCK 

#include <fitsio.h>
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

const bool WorkAroundBug = false; // set false to see bug, true to work around it

#ifdef HAVESYNCCLOCK
#include <syncclock.h>
#include "exposureTimer.h"
#endif
#include "pvcam/master.h"
#include "pvcam/pvcam.h"

// the following enum forces proper frame transfer only if pi133b.dat is present
// and now presumably pi133b2.dat & pi133b5.dat as well

#define PARAM_FORCE_READOUT_MODE ((CLASS2<<16) + (TYPE_UNS32<<24) + 326)

enum ForceReadOut {
    ALWAYS_CHECK_EXP,
    MAKE_FULL,
    MAKE_FRAME_TRANSFER,
    MAKE_AUTOMATIC
};

bool camOpen(int16 *hCamPtr);   // open the camera and set hCamPtr
void camClose(int16 *hCamPtr);  // close the camera and set hCamPtr invalid
bool camSetup(int16 hCam);  // Set up the CCD camera
bool getImages(int16 hCam);         // Take exposures
bool writeImage(uns16 *imPtr, int imNum);   // Write images as FITS files
void printCFitsIOError(int status); // Print CFITSIO Errors
void printPLError();                // Print PL (PVCAM) errors

const int ReadRateCode = 1; // 0=slow, 1=fast
const int GainCode = 2;     // 1=low, 2=medium, 3=high

const int CircBufImagesAlloc = 5;   // Number of images to allocate in circular buffer
const int CircBufImagesAdmitToPVCam = 5; // number of images to admit to PVCam
#define NamePrefixLen 50

// global exposure args
char namePrefix[NamePrefixLen] = "";
int numExp = 0;
double expTime = 0.0;
bool useExternalSync = false;
rgn_type pvRgn;

// computed globals
int16 hCam = -1;    // an invalid value, I hope
int serPix = 0;
int parPix = 0;

int main(int argc, char **argv) {
    if (!camOpen(&hCam)) goto errExit;

    strncpy(namePrefix, "1bias", NamePrefixLen);
    numExp = 3;
    expTime = 0.0;
    useExternalSync = false;
    pvRgn.s1 = 0;
    pvRgn.p1 = 0;
    pvRgn.s2 = 1023;
    pvRgn.p2 = 1023;
    pvRgn.sbin = 1;
    pvRgn.pbin = 1;
    if (!getImages(hCam)) goto errExit;

    if (WorkAroundBug) {
        camClose(&hCam);
        camOpen(&hCam);
    }

    strncpy(namePrefix, "2obj", NamePrefixLen);
    numExp = 3;
    expTime = 1.0;
    useExternalSync = true;
    pvRgn.s1 = 0;
    pvRgn.p1 = 0;
    pvRgn.s2 = 1023;
    pvRgn.p2 = 1023;
    pvRgn.sbin = 2;
    pvRgn.pbin = 2;
    if (!getImages(hCam)) goto errExit;


    if (WorkAroundBug) {
        camClose(&hCam);
        camOpen(&hCam);
    }

    strncpy(namePrefix, "3bias", NamePrefixLen);
    numExp = 3;
    expTime = 0.0;
    useExternalSync = false;
    pvRgn.s1 = 0;
    pvRgn.p1 = 0;
    pvRgn.s2 = 1023;
    pvRgn.p2 = 1023;
    pvRgn.sbin = 1;
    pvRgn.pbin = 1;
    if (!getImages(hCam)) goto errExit;

    if (WorkAroundBug) {
        camClose(&hCam);
        camOpen(&hCam);
    }

    strncpy(namePrefix, "4obj", NamePrefixLen);
    numExp = 3;
    expTime = 1.0;
    useExternalSync = true;
    pvRgn.s1 = 0;
    pvRgn.p1 = 0;
    pvRgn.s2 = 1023;
    pvRgn.p2 = 1023;
    pvRgn.sbin = 2;
    pvRgn.pbin = 2;
    if (!getImages(hCam)) goto errExit;

    camClose(&hCam);
    return 0;

errExit:
    camClose(&hCam);
    return 1;
}

/**
 * Open and configure the camera
 */
bool camOpen(int16 *hCamPtr) {
    char cam_name[CAM_NAME_LEN];
    *hCamPtr = -1;

    // this call will reliably fails with pl_error_code 0 
    // if the camera was already open (maybe one has to wait a few seconds? 1 is not enough)
    // but that appears to be harmless so just continue if error code is 0
    printf("pl_pvcam_init()\n");
    if (!pl_pvcam_init() && (pl_error_code() != 0)) goto plError;

    printf("pl_cam_get_name(0, cam_name)\n");
    if (!pl_cam_get_name(0, cam_name)) goto plError;
    printf("cam_name=%s\n", cam_name);

    printf("pl_cam_open(%s, hCamPtr, OPEN_EXCLUSIVE); this takes ~20 seconds\n", cam_name);
    if (!pl_cam_open(cam_name, hCamPtr, OPEN_EXCLUSIVE)) goto plError;
    printf("hCam=%d\n", *hCamPtr);
    
    if (!camSetup(*hCamPtr)) goto errorExit;

    return true;
plError:
    printPLError();
    goto errorExit;

errorExit:
    camClose(hCamPtr);
    return false;
}

/**
 * Close the camera
 */
void camClose(int16 *hCamPtr) {
    printf("Closing and uninitializing camera\n");
    if (*hCamPtr >= 0) {
        printf("pl_cam_close(%d)\n", *hCamPtr);
        pl_cam_close(*hCamPtr);
        *hCamPtr = -1;
    } else {
        printf("not calling pl_cam_close; hCam=%d is invalid\n", *hCamPtr);
    }

    printf("pl_pvcam_uninit\n");
    pl_pvcam_uninit();
    
}

/**
 * Configure the camera
 */
bool camSetup(int16 hCam) {    // setting up the camera 
    printf("Setting up the camera\n");

    // Configure CCD cooling and report
    const int16 desCCDTemp = -4000;   // Desired CCD temperature in 1/100 C
    int coolingMode = NORMAL_COOL;    // This corresponds to thermoelectric cooling
    int16 readCCDTemp;                // Read CCD temperature in 1/100 C
    rs_bool avail_flag;

    if (!pl_set_param(hCam, PARAM_COOLING_MODE, (void*)&coolingMode)) goto plError;

    if (!pl_set_param(hCam, PARAM_TEMP_SETPOINT, (void*)&desCCDTemp)) goto plError;

    if (!pl_get_param(hCam, PARAM_TEMP_SETPOINT, ATTR_CURRENT, &readCCDTemp)) goto plError;
    printf("\n\nCurrent CCD temperature setpoint: %4.1f C\n", (double)readCCDTemp/100);

    if (!pl_get_param(hCam, PARAM_TEMP, ATTR_CURRENT, &readCCDTemp)) goto plError;
    printf("Current CCD temperature: %4.1f\n", (double)readCCDTemp/100);

    // Setting the readout rate
    if (!pl_set_param(hCam, PARAM_SPDTAB_INDEX, (void*)&ReadRateCode)) goto plError;
    
    // Setting the amplifier gain
    if (!pl_set_param(hCam, PARAM_GAIN_INDEX, (void*)&GainCode)) goto plError;

    if (!pl_get_param(hCam, PARAM_CIRC_BUFFER, ATTR_AVAIL, &avail_flag)) goto plError;
    if (!avail_flag) {
        printf("Circular buffers not supported\n");
        return false;
    }

    int edgeTriggerMode = EDGE_TRIG_NEG;
    if (!pl_set_param(hCam, PARAM_EDGE_TRIGGER, (void*) &edgeTriggerMode)) goto plError;

    int forceReadoutMode = MAKE_FRAME_TRANSFER;
    if (!pl_set_param(hCam, PARAM_FORCE_READOUT_MODE, (void*) &forceReadoutMode)) goto plError;

    return true;
plError:
    printPLError();
    printf("camSetup failed\n");
    return false;
}

/**
 * Get a sequence of images
 *
 * Set the following global variables before calling:
 * * namePrefix
 * * numExp
 * * expTime
 * * useExternalSync
 * * pvRgn (fields are s1, p1, s2, p2, sbin, pbin)
 */
bool getImages(int16 hCam) {
    int imNum = 1;
    int numMissedExp = 0;
    uns16 *circBuffer = NULL;
    bool retVal = false;

    // flags for cleanup
    bool doUninitSeq = false;
    bool didSetupCont = false;
    bool doStopCont = false;
    bool doFinishSeq = false;
    bool doUnlockFrame = false;

    printf("\n\ngetimages\n");
    printf("Name Prefix=%s\n", namePrefix);
    printf("Number of images=%d\n", numExp);
    printf("Exposure time=%0.1f sec\n", expTime);
    printf("Image region: beg=%d, %d; end=%d, %d; bin=%d, %d\n\n",
        pvRgn.s1, pvRgn.p1, pvRgn.s2, pvRgn.p2, pvRgn.sbin, pvRgn.pbin);
    
    serPix = (1 + pvRgn.s2 - pvRgn.s1) / pvRgn.sbin;
    parPix = (1 + pvRgn.p2 - pvRgn.p1) / pvRgn.pbin;
    
    // Increase CCD size if necessary to accomodate overscan
    if ((pvRgn.s2 > 1023) || (pvRgn.p2 > 1023)) {
        rs_bool ON=1;
        if (!pl_set_param(hCam, PARAM_CUSTOM_CHIP, (void*)&ON)) goto plError;

        if (pvRgn.s2 > 1023) {
            uns16 serSize = pvRgn.s2 + 1;
            if (!pl_set_param(hCam, PARAM_SER_SIZE, (void*)&serSize)) goto plError;
        }
    
        if (pvRgn.p2 > 1023) {
            uns16 parSize = pvRgn.p2 + 1;
            if (!pl_set_param(hCam, PARAM_PAR_SIZE, (void*)&parSize)) goto plError;
        }
    }
    
    uns32 imBytes;                  // image size in bytes
    void_ptr imVoidPtr = NULL;
    int16 status, prevStatus;       // Readout status
    uns32 byteCnt, fullBufCnt;
    uns32 expTimeMS = (uns32) expTime * 1000;   // Exposure time in milliseconds
    int16 readCCDTemp;              // Read CCD temperature in 1/100 C

 
    flt64 readoutTimeMS;            // Readout time in milliseconds
    double actExpTime;
    struct timeval startTval;
    
    printf ("Setting up exposure read sequence\n");
    if (!pl_exp_init_seq()) goto plError;
    doUninitSeq = true;
    if (useExternalSync) {
        // using external sync pulse (Strobed mode)
        if (!pl_exp_setup_cont(hCam, 1, &pvRgn, STROBED_MODE, 0, &imBytes, CIRC_NO_OVERWRITE)) goto plError;
        didSetupCont = true;

        // Setting exposure time to 0 in strobed mode helps in allowing the sync pulse
        // to control not only the exposure start time, but also the exposure duration 
        // This will only work with negative edge/level triggering.
    } else {
        // using internal timer (Timed mode)
        if (!pl_exp_setup_cont(hCam, 1, &pvRgn, TIMED_MODE, expTimeMS, &imBytes, CIRC_NO_OVERWRITE)) goto plError;
        didSetupCont = true;
    }
    
    pl_get_param(hCam, PARAM_READOUT_TIME, ATTR_CURRENT, &readoutTimeMS);
    printf("Readout time = %.3f sec\n", readoutTimeMS / 1000.0);
    if ((expTime > 0) && (expTime < readoutTimeMS / 1000.0)) {
        printf("Error: exposure time %0.3f sec < readout time %0.3f sec\n",
            expTime, readoutTimeMS / 1000.0);
        goto nonPLError;
    }

    printf("Allocating circular buffer\n");
    circBuffer = (uns16*)malloc(imBytes * CircBufImagesAlloc);
    if (circBuffer == NULL) {
        printf("Could not allocate %ld bytes for %d-image circular buffer\n",
            (imBytes * CircBufImagesAlloc), CircBufImagesAlloc);
        goto nonPLError;
    }

    printf("Starting exposure read sequence\n");
    if (!pl_exp_start_cont(hCam, circBuffer, imBytes*CircBufImagesAdmitToPVCam)) goto plError;
    doStopCont = true;
    doFinishSeq = true;
    
#ifdef HAVESYNCCLOCK
    if (useExternalSync) {
        // start the sync pulse
        if (setExpTime(&actExpTime, &startTval, expTime) < 0) {
            printf("Could not start the sync pulse counter\n");
            goto nonPLError;
        }
        printf("Started sync pulse counter with interval %.2f\n", actExpTime);
    }
#endif

    if (useExternalSync) {
        printf("\nBeginning Image Acquisition with external sync timing\n\n");
    } else {
        printf("\nBeginning Image Acquisition with internal camera timing\n\n");
    }

    // Acquisition loop
    while(imNum <= numExp) {
        printf("Acquiring image %d of %d\n", imNum, numExp);
        
        // wait for data or error
        prevStatus = 0;
        while(true) {
            if (!pl_exp_check_cont_status(hCam, &status, &byteCnt, &fullBufCnt)) {
                printf("Exposure sequence failed: ");
                goto plError;
            }

            if (status == READOUT_COMPLETE) {
                // data is available
                printf("Exposure available\n");
                break;
            } else if (status == READOUT_FAILED) {
                printf("Readout failed: ");
                goto plError;
            } else if (status != prevStatus) {
                printf("Exposure status = %d\n", status);
            }
            prevStatus = status;   
        }
        if (!pl_exp_get_oldest_frame(hCam, &imVoidPtr)) {
            printf("Could not retrieve image from circular buffer: ");
            goto plError;
        }
        doUnlockFrame = true;
        uns16 *imPtr = (uns16*)imVoidPtr;

        unsigned long int imPixels = serPix * parPix;
        uns16 endPixel = *(imPtr + imPixels - 1);
        printf("Acquired an image; center pixel=%u; end pixel=%u\n", *(imPtr + (imPixels/2) - 1), endPixel);
        if (endPixel == 0) printf("*** WARNING *** end pixel is zero\n");

        if (!pl_get_param(hCam, PARAM_TEMP, ATTR_CURRENT, &readCCDTemp)) goto plError;
        printf("Current CCD temperature: %4.1f\n", (double)readCCDTemp/100);
                                                                                                               
        if (!writeImage(imPtr, imNum)) goto nonPLError;
        
        if (doUnlockFrame) {
            doUnlockFrame = false; // if call fails, don't try again
            if (!pl_exp_unlock_oldest_frame(hCam)) goto plError;
        }

        imNum += 1;
    } // end while 

    retVal = true;
    goto cleanup;

nonPLError:
    retVal = false;
    goto cleanup;

plError:
    retVal = false;
    printPLError();
    goto cleanup;

cleanup:
    if (numMissedExp > 0) {
        printf("WARNING: missed a total of %d exposures!\n", numMissedExp);
    }
    if (doUnlockFrame) {
        pl_exp_unlock_oldest_frame(hCam);
    }
    if (circBuffer != NULL) {
        free(circBuffer);
    }
    if (doStopCont) {
        pl_exp_stop_cont(hCam, CCS_HALT);           // Stop the continuous acquisition
                                                    // Camera Control Sub-system halt
    }
    if (doFinishSeq) {
        pl_exp_finish_seq(hCam, circBuffer, 0);     // Finish the sequence
    }
    if (doUninitSeq) {
        pl_exp_uninit_seq();
    }
    return retVal;
}


/**
 * Save an image to a FITS file
 */
bool writeImage(uns16 *imPtr, int imNum) {
    fitsfile *fptr;       // pointer to the FITS file, defined in fitsio.h
    int status = 0;       // initialize status before calling fitsio routines
    long first_pixel = 1, nelements;

    // initialize FITS image parameters
    bool fileOpen = false;
    char filename[265];
    int bitpix   =  USHORT_IMG;     // 16-bit unsigned short pixel values
    long naxis    = 2;              // 2-dimensional image     
    long naxes[2] = {serPix, parPix};   // image is serPix times parPix pixels

    sprintf(filename,"%s.%03d.fits",namePrefix,imNum);

    if (fits_create_file(&fptr, filename, &status)) goto errExit;
    fileOpen = true;

    if (fits_create_img(fptr, bitpix, naxis, naxes, &status)) goto errExit;          

    nelements = naxes[0] * naxes[1];          // number of pixels to write

    // write the array of unsigned integers to the FITS file
    if (fits_write_img(fptr, TUSHORT, first_pixel, nelements, imPtr, &status)) goto errExit;
      
    if (fits_close_file(fptr, &status)) goto errExit;           

    printf("Saved image %s\n", filename);

    return true;

errExit: // status should be set
    printCFitsIOError(status);
    if (fileOpen) {
        if (fits_close_file(fptr, &status)) {
            printCFitsIOError(status);
        }
    }
    return false;
}


/*****************************************************/
// Print out cfitsio error messages and exit program
/*****************************************************/
void printCFitsIOError(int status) {
    if (status) {
       fits_report_error(stderr, status); // print error report
    }
    return;
}


/* Print error message associated with the most recent PVCam error.
Does nothing if there is no error.
*/
void printPLError() {
    int16 plerrcode;
    char plerrmsg[ERROR_MSG_LEN];

    plerrcode = pl_error_code();
    
    if (!pl_error_message(plerrcode, plerrmsg)) {
        snprintf(plerrmsg, ERROR_MSG_LEN, "PVCam error %d", plerrcode);
    }
    printf("%s\n", plerrmsg);
}
